# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.model import from_disk


class StorageData:
    def __getattr__(self, key):
        try:
            v = globals()[key]
        except KeyError as e:
            raise AttributeError(e.args[0])
        if hasattr(v, "copy"):
            return v.copy()
        return v


data = StorageData()


cont = {
    "data": b"42\n",
    "length": 3,
    "sha1": hash_to_bytes("34973274ccef6ab4dfaaf86599792fa9c3fe4689"),
    "sha1_git": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
    "sha256": hash_to_bytes(
        "673650f936cb3b0a2f93ce09d81be10748b1b203c19e8176b4eefc1964a0cf3a"
    ),
    "blake2s256": hash_to_bytes(
        "d5fe1939576527e42cfd76a9455a2432fe7f56669564577dd93c4280e76d661d"
    ),
    "status": "visible",
}

cont2 = {
    "data": b"4242\n",
    "length": 5,
    "sha1": hash_to_bytes("61c2b3a30496d329e21af70dd2d7e097046d07b7"),
    "sha1_git": hash_to_bytes("36fade77193cb6d2bd826161a0979d64c28ab4fa"),
    "sha256": hash_to_bytes(
        "859f0b154fdb2d630f45e1ecae4a862915435e663248bb8461d914696fc047cd"
    ),
    "blake2s256": hash_to_bytes(
        "849c20fad132b7c2d62c15de310adfe87be94a379941bed295e8141c6219810d"
    ),
    "status": "visible",
}

cont3 = {
    "data": b"424242\n",
    "length": 7,
    "sha1": hash_to_bytes("3e21cc4942a4234c9e5edd8a9cacd1670fe59f13"),
    "sha1_git": hash_to_bytes("c932c7649c6dfa4b82327d121215116909eb3bea"),
    "sha256": hash_to_bytes(
        "92fb72daf8c6818288a35137b72155f507e5de8d892712ab96277aaed8cf8a36"
    ),
    "blake2s256": hash_to_bytes(
        "76d0346f44e5a27f6bafdd9c2befd304aff83780f93121d801ab6a1d4769db11"
    ),
    "status": "visible",
    "ctime": "2019-12-01 00:00:00Z",
}

contents = (cont, cont2, cont3)


missing_cont = {
    "data": b"missing\n",
    "length": 8,
    "sha1": hash_to_bytes("f9c24e2abb82063a3ba2c44efd2d3c797f28ac90"),
    "sha1_git": hash_to_bytes("33e45d56f88993aae6a0198013efa80716fd8919"),
    "sha256": hash_to_bytes(
        "6bbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a"
    ),
    "blake2s256": hash_to_bytes(
        "306856b8fd879edb7b6f1aeaaf8db9bbecc993cd7f776c333ac3a782fa5c6eba"
    ),
    "status": "absent",
}

skipped_cont = {
    "length": 1024 * 1024 * 200,
    "sha1_git": hash_to_bytes("33e45d56f88993aae6a0198013efa80716fd8920"),
    "sha1": hash_to_bytes("43e45d56f88993aae6a0198013efa80716fd8920"),
    "sha256": hash_to_bytes(
        "7bbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a"
    ),
    "blake2s256": hash_to_bytes(
        "ade18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b"
    ),
    "reason": "Content too long",
    "status": "absent",
    "origin": "file:///dev/zero",
}

skipped_cont2 = {
    "length": 1024 * 1024 * 300,
    "sha1_git": hash_to_bytes("44e45d56f88993aae6a0198013efa80716fd8921"),
    "sha1": hash_to_bytes("54e45d56f88993aae6a0198013efa80716fd8920"),
    "sha256": hash_to_bytes(
        "8cbd052ab054ef222c1c87be60cd191addedd24cc882d1f5f7f7be61dc61bb3a"
    ),
    "blake2s256": hash_to_bytes(
        "9ce18b1adecb33f891ca36664da676e12c772cc193778aac9a137b8dc5834b9b"
    ),
    "reason": "Content too long",
    "status": "absent",
}

skipped_contents = (skipped_cont, skipped_cont2)


dir = {
    "id": hash_to_bytes("34f335a750111ca0a8b64d8034faec9eedc396be"),
    "entries": (
        {
            "name": b"foo",
            "type": "file",
            "target": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),  # cont
            "perms": from_disk.DentryPerms.content,
        },
        {
            "name": b"bar\xc3",
            "type": "dir",
            "target": b"12345678901234567890",
            "perms": from_disk.DentryPerms.directory,
        },
    ),
}

dir2 = {
    "id": hash_to_bytes("8505808532953da7d2581741f01b29c04b1cb9ab"),
    "entries": (
        {
            "name": b"oof",
            "type": "file",
            "target": hash_to_bytes(  # cont2
                "36fade77193cb6d2bd826161a0979d64c28ab4fa"
            ),
            "perms": from_disk.DentryPerms.content,
        },
    ),
}

dir3 = {
    "id": hash_to_bytes("4ea8c6b2f54445e5dd1a9d5bb2afd875d66f3150"),
    "entries": (
        {
            "name": b"foo",
            "type": "file",
            "target": hash_to_bytes("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),  # cont
            "perms": from_disk.DentryPerms.content,
        },
        {
            "name": b"subdir",
            "type": "dir",
            "target": hash_to_bytes("34f335a750111ca0a8b64d8034faec9eedc396be"),  # dir
            "perms": from_disk.DentryPerms.directory,
        },
        {
            "name": b"hello",
            "type": "file",
            "target": b"12345678901234567890",
            "perms": from_disk.DentryPerms.content,
        },
    ),
}

dir4 = {
    "id": hash_to_bytes("377aa5fcd944fbabf502dbfda55cd14d33c8c3c6"),
    "entries": (
        {
            "name": b"subdir1",
            "type": "dir",
            "target": hash_to_bytes("4ea8c6b2f54445e5dd1a9d5bb2afd875d66f3150"),  # dir3
            "perms": from_disk.DentryPerms.directory,
        },
    ),
}

directories = (dir, dir2, dir3, dir4)


minus_offset = datetime.timezone(datetime.timedelta(minutes=-120))
plus_offset = datetime.timezone(datetime.timedelta(minutes=120))

revision = {
    "id": hash_to_bytes("066b1b62dbfa033362092af468bf6cfabec230e7"),
    "message": b"hello",
    "author": {
        "name": b"Nicolas Dandrimont",
        "email": b"nicolas@example.com",
        "fullname": b"Nicolas Dandrimont <nicolas@example.com> ",
    },
    "date": {
        "timestamp": {"seconds": 1234567890, "microseconds": 0},
        "offset": 120,
        "negative_utc": False,
    },
    "committer": {
        "name": b"St\xc3fano Zacchiroli",
        "email": b"stefano@example.com",
        "fullname": b"St\xc3fano Zacchiroli <stefano@example.com>",
    },
    "committer_date": {
        "timestamp": {"seconds": 1123456789, "microseconds": 0},
        "offset": 0,
        "negative_utc": True,
    },
    "parents": (b"01234567890123456789", b"23434512345123456789"),
    "type": "git",
    "directory": hash_to_bytes("34f335a750111ca0a8b64d8034faec9eedc396be"),  # dir
    "metadata": {
        "checksums": {"sha1": "tarball-sha1", "sha256": "tarball-sha256",},
        "signed-off-by": "some-dude",
        "extra_headers": [
            ["gpgsig", b"test123"],
            ["mergetag", b"foo\\bar"],
            ["mergetag", b"\x22\xaf\x89\x80\x01\x00"],
        ],
    },
    "synthetic": True,
}

revision2 = {
    "id": hash_to_bytes("df7a6f6a99671fb7f7343641aff983a314ef6161"),
    "message": b"hello again",
    "author": {
        "name": b"Roberto Dicosmo",
        "email": b"roberto@example.com",
        "fullname": b"Roberto Dicosmo <roberto@example.com>",
    },
    "date": {
        "timestamp": {"seconds": 1234567843, "microseconds": 220000,},
        "offset": -720,
        "negative_utc": False,
    },
    "committer": {
        "name": b"tony",
        "email": b"ar@dumont.fr",
        "fullname": b"tony <ar@dumont.fr>",
    },
    "committer_date": {
        "timestamp": {"seconds": 1123456789, "microseconds": 0},
        "offset": 0,
        "negative_utc": False,
    },
    "parents": (b"01234567890123456789",),
    "type": "git",
    "directory": hash_to_bytes("8505808532953da7d2581741f01b29c04b1cb9ab"),  # dir2
    "metadata": None,
    "synthetic": False,
}

revision3 = {
    "id": hash_to_bytes("2cbd7bb22c653bbb23a29657852a50a01b591d46"),
    "message": b"a simple revision with no parents this time",
    "author": {
        "name": b"Roberto Dicosmo",
        "email": b"roberto@example.com",
        "fullname": b"Roberto Dicosmo <roberto@example.com>",
    },
    "date": {
        "timestamp": {"seconds": 1234567843, "microseconds": 220000,},
        "offset": -720,
        "negative_utc": False,
    },
    "committer": {
        "name": b"tony",
        "email": b"ar@dumont.fr",
        "fullname": b"tony <ar@dumont.fr>",
    },
    "committer_date": {
        "timestamp": {"seconds": 1127351742, "microseconds": 0},
        "offset": 0,
        "negative_utc": False,
    },
    "parents": (),
    "type": "git",
    "directory": hash_to_bytes("8505808532953da7d2581741f01b29c04b1cb9ab"),  # dir2
    "metadata": None,
    "synthetic": True,
}

revision4 = {
    "id": hash_to_bytes("88cd5126fc958ed70089d5340441a1c2477bcc20"),
    "message": b"parent of self.revision2",
    "author": {
        "name": b"me",
        "email": b"me@soft.heri",
        "fullname": b"me <me@soft.heri>",
    },
    "date": {
        "timestamp": {"seconds": 1244567843, "microseconds": 220000,},
        "offset": -720,
        "negative_utc": False,
    },
    "committer": {
        "name": b"committer-dude",
        "email": b"committer@dude.com",
        "fullname": b"committer-dude <committer@dude.com>",
    },
    "committer_date": {
        "timestamp": {"seconds": 1244567843, "microseconds": 220000,},
        "offset": -720,
        "negative_utc": False,
    },
    "parents": (
        hash_to_bytes("2cbd7bb22c653bbb23a29657852a50a01b591d46"),
    ),  # revision3
    "type": "git",
    "directory": hash_to_bytes("34f335a750111ca0a8b64d8034faec9eedc396be"),  # dir
    "metadata": None,
    "synthetic": False,
}

revisions = (revision, revision2, revision3, revision4)


origin = {
    "url": "file:///dev/null",
}

origin2 = {
    "url": "file:///dev/zero",
}

origins = (origin, origin2)


metadata_authority = {
    "type": "deposit",
    "url": "http://hal.inria.example.com/",
    "metadata": {"location": "France"},
}
metadata_authority2 = {
    "type": "registry",
    "url": "http://wikidata.example.com/",
    "metadata": {},
}

metadata_fetcher = {
    "name": "swh-deposit",
    "version": "0.0.1",
    "metadata": {"sword_version": "2"},
}
metadata_fetcher2 = {
    "name": "swh-example",
    "version": "0.0.1",
    "metadata": {},
}

date_visit1 = datetime.datetime(2015, 1, 1, 23, 0, 0, tzinfo=datetime.timezone.utc)
type_visit1 = "git"

date_visit2 = datetime.datetime(2017, 1, 1, 23, 0, 0, tzinfo=datetime.timezone.utc)
type_visit2 = "hg"

date_visit3 = datetime.datetime(2018, 1, 1, 23, 0, 0, tzinfo=datetime.timezone.utc)
type_visit3 = "deb"

release = {
    "id": hash_to_bytes("a673e617fcc6234e29b2cad06b8245f96c415c61"),
    "name": b"v0.0.1",
    "author": {
        "name": b"olasd",
        "email": b"nic@olasd.fr",
        "fullname": b"olasd <nic@olasd.fr>",
    },
    "date": {
        "timestamp": {"seconds": 1234567890, "microseconds": 0},
        "offset": 42,
        "negative_utc": False,
    },
    "target": b"43210987654321098765",
    "target_type": "revision",
    "message": b"synthetic release",
    "synthetic": True,
}

release2 = {
    "id": hash_to_bytes("6902bd4c82b7d19a421d224aedab2b74197e420d"),
    "name": b"v0.0.2",
    "author": {
        "name": b"tony",
        "email": b"ar@dumont.fr",
        "fullname": b"tony <ar@dumont.fr>",
    },
    "date": {
        "timestamp": {"seconds": 1634366813, "microseconds": 0},
        "offset": -120,
        "negative_utc": False,
    },
    "target": b"432109\xa9765432\xc309\x00765",
    "target_type": "revision",
    "message": b"v0.0.2\nMisc performance improvements + bug fixes",
    "synthetic": False,
}

release3 = {
    "id": hash_to_bytes("3e9050196aa288264f2a9d279d6abab8b158448b"),
    "name": b"v0.0.2",
    "author": {
        "name": b"tony",
        "email": b"tony@ardumont.fr",
        "fullname": b"tony <tony@ardumont.fr>",
    },
    "date": {
        "timestamp": {"seconds": 1634336813, "microseconds": 0},
        "offset": 0,
        "negative_utc": False,
    },
    "target": hash_to_bytes("df7a6f6a99671fb7f7343641aff983a314ef6161"),
    "target_type": "revision",
    "message": b"yet another synthetic release",
    "synthetic": True,
}

releases = (release, release2, release3)


snapshot = {
    "id": hash_to_bytes("409ee1ff3f10d166714bc90581debfd0446dda57"),
    "branches": {
        b"master": {
            "target": hash_to_bytes("066b1b62dbfa033362092af468bf6cfabec230e7"),
            "target_type": "revision",
        },
    },
}

empty_snapshot = {
    "id": hash_to_bytes("1a8893e6a86f444e8be8e7bda6cb34fb1735a00e"),
    "branches": {},
}

complete_snapshot = {
    "id": hash_to_bytes("a56ce2d81c190023bb99a3a36279307522cb85f6"),
    "branches": {
        b"directory": {
            "target": hash_to_bytes("1bd0e65f7d2ff14ae994de17a1e7fe65111dcad8"),
            "target_type": "directory",
        },
        b"directory2": {
            "target": hash_to_bytes("1bd0e65f7d2ff14ae994de17a1e7fe65111dcad8"),
            "target_type": "directory",
        },
        b"content": {
            "target": hash_to_bytes("fe95a46679d128ff167b7c55df5d02356c5a1ae1"),
            "target_type": "content",
        },
        b"alias": {"target": b"revision", "target_type": "alias",},
        b"revision": {
            "target": hash_to_bytes("aafb16d69fd30ff58afdd69036a26047f3aebdc6"),
            "target_type": "revision",
        },
        b"release": {
            "target": hash_to_bytes("7045404f3d1c54e6473c71bbb716529fbad4be24"),
            "target_type": "release",
        },
        b"snapshot": {
            "target": hash_to_bytes("1a8893e6a86f444e8be8e7bda6cb34fb1735a00e"),
            "target_type": "snapshot",
        },
        b"dangling": None,
    },
}

snapshots = (snapshot, empty_snapshot, complete_snapshot)

content_metadata = {
    "id": f"swh:1:cnt:{cont['sha1_git']}",
    "context": {"origin": origin["url"]},
    "discovery_date": datetime.datetime(
        2015, 1, 1, 21, 0, 0, tzinfo=datetime.timezone.utc
    ),
    "authority": {
        "type": metadata_authority["type"],
        "url": metadata_authority["url"],
    },
    "fetcher": {
        "name": metadata_fetcher["name"],
        "version": metadata_fetcher["version"],
    },
    "format": "json",
    "metadata": b'{"foo": "bar"}',
}
content_metadata2 = {
    "id": f"swh:1:cnt:{cont['sha1_git']}",
    "context": {"origin": origin2["url"]},
    "discovery_date": datetime.datetime(
        2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
    ),
    "authority": {
        "type": metadata_authority["type"],
        "url": metadata_authority["url"],
    },
    "fetcher": {
        "name": metadata_fetcher["name"],
        "version": metadata_fetcher["version"],
    },
    "format": "yaml",
    "metadata": b"foo: bar",
}
content_metadata3 = {
    "id": f"swh:1:cnt:{cont['sha1_git']}",
    "context": {
        "origin": origin["url"],
        "visit": 42,
        "snapshot": f"swh:1:snp:{hash_to_hex(snapshot['id'])}",
        "release": f"swh:1:rel:{hash_to_hex(release['id'])}",
        "revision": f"swh:1:rev:{hash_to_hex(revision['id'])}",
        "directory": f"swh:1:dir:{hash_to_hex(dir['id'])}",
        "path": b"/foo/bar",
    },
    "discovery_date": datetime.datetime(
        2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
    ),
    "authority": {
        "type": metadata_authority2["type"],
        "url": metadata_authority2["url"],
    },
    "fetcher": {
        "name": metadata_fetcher2["name"],
        "version": metadata_fetcher2["version"],
    },
    "format": "yaml",
    "metadata": b"foo: bar",
}

origin_metadata = {
    "origin_url": origin["url"],
    "discovery_date": datetime.datetime(
        2015, 1, 1, 21, 0, 0, tzinfo=datetime.timezone.utc
    ),
    "authority": {
        "type": metadata_authority["type"],
        "url": metadata_authority["url"],
    },
    "fetcher": {
        "name": metadata_fetcher["name"],
        "version": metadata_fetcher["version"],
    },
    "format": "json",
    "metadata": b'{"foo": "bar"}',
}
origin_metadata2 = {
    "origin_url": origin["url"],
    "discovery_date": datetime.datetime(
        2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
    ),
    "authority": {
        "type": metadata_authority["type"],
        "url": metadata_authority["url"],
    },
    "fetcher": {
        "name": metadata_fetcher["name"],
        "version": metadata_fetcher["version"],
    },
    "format": "yaml",
    "metadata": b"foo: bar",
}
origin_metadata3 = {
    "origin_url": origin["url"],
    "discovery_date": datetime.datetime(
        2017, 1, 1, 22, 0, 0, tzinfo=datetime.timezone.utc
    ),
    "authority": {
        "type": metadata_authority2["type"],
        "url": metadata_authority2["url"],
    },
    "fetcher": {
        "name": metadata_fetcher2["name"],
        "version": metadata_fetcher2["version"],
    },
    "format": "yaml",
    "metadata": b"foo: bar",
}

person = {
    "name": b"John Doe",
    "email": b"john.doe@institute.org",
    "fullname": b"John Doe <john.doe@institute.org>",
}

objects = {
    "content": contents,
    "skipped_content": skipped_contents,
    "directory": directories,
    "revision": revisions,
    "origin": origins,
    "release": releases,
    "snapshot": snapshots,
}



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from unittest.mock import patch

from swh.model.model import Origin, OriginVisit, OriginVisitStatus, Snapshot

from swh.storage.algos.origin import iter_origins, origin_get_latest_visit_status
from swh.storage.utils import now

from swh.storage.tests.test_storage import round_to_milliseconds
from swh.storage.tests.storage_data import data


def assert_list_eq(left, right, msg=None):
    assert list(left) == list(right), msg


@pytest.fixture
def swh_storage_backend_config():
    yield {"cls": "validate", "storage": {"cls": "memory",}}


def test_iter_origins(swh_storage):
    origins = [
        {"url": "bar"},
        {"url": "qux"},
        {"url": "quuz"},
    ]
    assert swh_storage.origin_add(origins) == {"origin:add": 3}
    assert_list_eq(iter_origins(swh_storage), origins)
    assert_list_eq(iter_origins(swh_storage, batch_size=1), origins)
    assert_list_eq(iter_origins(swh_storage, batch_size=2), origins)

    for i in range(1, 5):
        assert_list_eq(iter_origins(swh_storage, origin_from=i + 1), origins[i:], i)

        assert_list_eq(
            iter_origins(swh_storage, origin_from=i + 1, batch_size=1), origins[i:], i
        )

        assert_list_eq(
            iter_origins(swh_storage, origin_from=i + 1, batch_size=2), origins[i:], i
        )

        for j in range(i, 5):
            assert_list_eq(
                iter_origins(swh_storage, origin_from=i + 1, origin_to=j + 1),
                origins[i:j],
                (i, j),
            )

            assert_list_eq(
                iter_origins(
                    swh_storage, origin_from=i + 1, origin_to=j + 1, batch_size=1
                ),
                origins[i:j],
                (i, j),
            )

            assert_list_eq(
                iter_origins(
                    swh_storage, origin_from=i + 1, origin_to=j + 1, batch_size=2
                ),
                origins[i:j],
                (i, j),
            )


@patch("swh.storage.in_memory.InMemoryStorage.origin_get_range")
def test_iter_origins_batch_size(mock_origin_get_range, swh_storage):
    mock_origin_get_range.return_value = []

    list(iter_origins(swh_storage))
    mock_origin_get_range.assert_called_with(origin_from=1, origin_count=10000)

    list(iter_origins(swh_storage, batch_size=42))
    mock_origin_get_range.assert_called_with(origin_from=1, origin_count=42)


def test_origin_get_latest_visit_status_none(swh_storage):
    """Looking up unknown objects should return nothing

    """
    # unknown origin so no result
    assert origin_get_latest_visit_status(swh_storage, "unknown-origin") is None

    # unknown type so no result
    origin = Origin.from_dict(data.origin)
    swh_storage.origin_add_one(origin)
    swh_storage.origin_visit_add(
        [OriginVisit(origin=origin.url, date=data.date_visit1, type="git",),]
    )[0]
    actual_origin_visit = origin_get_latest_visit_status(
        swh_storage, origin.url, type="unknown"
    )
    assert actual_origin_visit is None

    actual_origin_visit = origin_get_latest_visit_status(
        swh_storage, origin.url, require_snapshot=True
    )
    assert actual_origin_visit is None

    actual_origin_visit = origin_get_latest_visit_status(
        swh_storage, origin.url, allowed_statuses=["unknown"]
    )
    assert actual_origin_visit is None


def init_storage_with_origin_visits(swh_storage):
    """Initialize storage with origin/origin-visit/origin-visit-status

    """
    origin1 = Origin.from_dict(data.origin)
    origin2 = Origin.from_dict(data.origin2)
    swh_storage.origin_add([origin1, origin2])

    ov1, ov2 = swh_storage.origin_visit_add(
        [
            OriginVisit(
                origin=origin1.url, date=data.date_visit1, type=data.type_visit1,
            ),
            OriginVisit(
                origin=origin2.url, date=data.date_visit2, type=data.type_visit2,
            ),
        ]
    )

    snapshot = Snapshot.from_dict(data.complete_snapshot)
    swh_storage.snapshot_add([snapshot])

    date_now = now()
    date_now = round_to_milliseconds(date_now)
    assert data.date_visit1 < data.date_visit2
    assert data.date_visit2 < date_now

    # origin visit status 1 for origin visit 1
    ovs11 = OriginVisitStatus(
        origin=origin1.url,
        visit=ov1.visit,
        date=data.date_visit1,
        status="partial",
        snapshot=None,
    )
    # origin visit status 2 for origin visit 1
    ovs12 = OriginVisitStatus(
        origin=origin1.url,
        visit=ov1.visit,
        date=data.date_visit2,
        status="ongoing",
        snapshot=None,
    )
    # origin visit status 1 for origin visit 2
    ovs21 = OriginVisitStatus(
        origin=origin2.url,
        visit=ov2.visit,
        date=data.date_visit2,
        status="ongoing",
        snapshot=None,
    )
    # origin visit status 2 for origin visit 2
    ovs22 = OriginVisitStatus(
        origin=origin2.url,
        visit=ov2.visit,
        date=date_now,
        status="full",
        snapshot=snapshot.id,
        metadata={"something": "wicked"},
    )

    swh_storage.origin_visit_status_add([ovs11, ovs12, ovs21, ovs22])
    return {
        "origin": [origin1, origin2],
        "origin_visit": [ov1, ov2],
        "origin_visit_status": [ovs11, ovs12, ovs21, ovs22],
    }


def test_origin_get_latest_visit_status_filter_type(swh_storage):
    """Filtering origin visit per types should yield consistent results

    """
    objects = init_storage_with_origin_visits(swh_storage)
    origin1, origin2 = objects["origin"]
    ov1, ov2 = objects["origin_visit"]
    ovs11, ovs12, _, ovs22 = objects["origin_visit_status"]

    # no visit for origin1 url with type_visit2
    assert (
        origin_get_latest_visit_status(swh_storage, origin1.url, type=data.type_visit2)
        is None
    )

    # no visit for origin2 url with type_visit1
    assert (
        origin_get_latest_visit_status(swh_storage, origin2.url, type=data.type_visit1)
        is None
    )

    # Two visits, both with no snapshot, take the most recent
    actual_ov1, actual_ovs12 = origin_get_latest_visit_status(
        swh_storage, origin1.url, type=data.type_visit1
    )
    assert isinstance(actual_ov1, OriginVisit)
    assert isinstance(actual_ovs12, OriginVisitStatus)
    assert actual_ov1.origin == ov1.origin
    assert actual_ov1.visit == ov1.visit
    assert actual_ov1.type == data.type_visit1
    assert actual_ovs12 == ovs12

    # take the most recent visit with type_visit2
    actual_ov2, actual_ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, type=data.type_visit2
    )
    assert isinstance(actual_ov2, OriginVisit)
    assert isinstance(actual_ovs22, OriginVisitStatus)
    assert actual_ov2.origin == ov2.origin
    assert actual_ov2.visit == ov2.visit
    assert actual_ov2.type == data.type_visit2
    assert actual_ovs22 == ovs22


def test_origin_get_latest_visit_status_filter_status(swh_storage):
    objects = init_storage_with_origin_visits(swh_storage)
    origin1, origin2 = objects["origin"]
    ov1, ov2 = objects["origin_visit"]
    ovs11, ovs12, _, ovs22 = objects["origin_visit_status"]

    # no failed status for that visit
    assert (
        origin_get_latest_visit_status(
            swh_storage, origin2.url, allowed_statuses=["failed"]
        )
        is None
    )

    # only 1 partial for that visit
    actual_ov1, actual_ovs11 = origin_get_latest_visit_status(
        swh_storage, origin1.url, allowed_statuses=["partial"]
    )
    assert actual_ov1.origin == ov1.origin
    assert actual_ov1.visit == ov1.visit
    assert actual_ov1.type == data.type_visit1
    assert actual_ovs11 == ovs11

    # both status exist, take the latest one
    actual_ov1, actual_ovs12 = origin_get_latest_visit_status(
        swh_storage, origin1.url, allowed_statuses=["partial", "ongoing"]
    )
    assert actual_ov1.origin == ov1.origin
    assert actual_ov1.visit == ov1.visit
    assert actual_ov1.type == data.type_visit1
    assert actual_ovs12 == ovs12

    assert isinstance(actual_ov1, OriginVisit)
    assert isinstance(actual_ovs12, OriginVisitStatus)
    assert actual_ov1.origin == ov1.origin
    assert actual_ov1.visit == ov1.visit
    assert actual_ov1.type == data.type_visit1
    assert actual_ovs12 == ovs12

    # take the most recent visit with type_visit2
    actual_ov2, actual_ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, allowed_statuses=["full"]
    )
    assert actual_ov2.origin == ov2.origin
    assert actual_ov2.visit == ov2.visit
    assert actual_ov2.type == data.type_visit2
    assert actual_ovs22 == ovs22


def test_origin_get_latest_visit_status_filter_snapshot(swh_storage):
    objects = init_storage_with_origin_visits(swh_storage)
    origin1, origin2 = objects["origin"]
    _, ov2 = objects["origin_visit"]
    _, _, _, ovs22 = objects["origin_visit_status"]

    # there is no visit with snapshot yet for that visit
    assert (
        origin_get_latest_visit_status(swh_storage, origin1.url, require_snapshot=True)
        is None
    )

    # visit status with partial status visit elected
    actual_ov2, actual_ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, require_snapshot=True
    )
    assert actual_ov2.origin == ov2.origin
    assert actual_ov2.visit == ov2.visit
    assert actual_ov2.type == ov2.type
    assert actual_ovs22 == ovs22

    date_now = now()

    # Add another visit
    swh_storage.origin_visit_add(
        [OriginVisit(origin=origin2.url, date=date_now, type=data.type_visit2,),]
    )

    # Requiring the latest visit with a snapshot, we still find the previous visit
    ov2, ovs22 = origin_get_latest_visit_status(
        swh_storage, origin2.url, require_snapshot=True
    )
    assert actual_ov2.origin == ov2.origin
    assert actual_ov2.visit == ov2.visit
    assert actual_ov2.type == ov2.type
    assert actual_ovs22 == ovs22



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis.strategies import binary, composite, sets

from swh.model.hashutil import MultiHash


def gen_raw_content():
    """Generate raw content binary.

    """
    return binary(min_size=20, max_size=100)


@composite
def gen_contents(draw, *, min_size=0, max_size=100):
    """Generate valid and consistent content.

    Context: Test purposes

    Args:
        **draw**: Used by hypothesis to generate data
        **min_size** (int): Minimal number of elements to generate
                            (default: 0)
        **max_size** (int): Maximal number of elements to generate
                            (default: 100)

    Returns:
        [dict] representing contents. The list's size is between
        [min_size:max_size].

    """
    raw_contents = draw(sets(gen_raw_content(), min_size=min_size, max_size=max_size))

    contents = []
    for raw_content in raw_contents:
        contents.append(
            {
                "data": raw_content,
                "length": len(raw_content),
                "status": "visible",
                **MultiHash.from_data(raw_content).digest(),
            }
        )

    return contents



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given
import pytest

from swh.model.hypothesis_strategies import snapshots, branch_names, branch_targets
from swh.model.identifiers import snapshot_identifier, identifier_to_bytes
from swh.model.model import Origin, OriginVisit, OriginVisitStatus, Snapshot

from swh.storage.algos.snapshot import snapshot_get_all_branches, snapshot_get_latest
from swh.storage.utils import now

from swh.storage.tests.storage_data import data


@pytest.fixture
def swh_storage_backend_config():
    yield {
        "cls": "memory",
        "journal_writer": None,
    }


@given(snapshot=snapshots(min_size=0, max_size=10, only_objects=False))
def test_snapshot_small(swh_storage, snapshot):  # noqa
    snapshot = snapshot.to_dict()
    swh_storage.snapshot_add([snapshot])

    returned_snapshot = snapshot_get_all_branches(swh_storage, snapshot["id"])
    assert snapshot == returned_snapshot


@given(branch_name=branch_names(), branch_target=branch_targets(only_objects=True))
def test_snapshot_large(swh_storage, branch_name, branch_target):  # noqa
    branch_target = branch_target.to_dict()

    snapshot = {
        "branches": {b"%s%05d" % (branch_name, i): branch_target for i in range(10000)}
    }
    snapshot["id"] = identifier_to_bytes(snapshot_identifier(snapshot))

    swh_storage.snapshot_add([snapshot])

    returned_snapshot = snapshot_get_all_branches(swh_storage, snapshot["id"])
    assert snapshot == returned_snapshot


def test_snapshot_get_latest_none(swh_storage):
    """Retrieve latest snapshot on unknown origin or origin without snapshot should
    yield no result

    """
    # unknown origin so None
    assert snapshot_get_latest(swh_storage, "unknown-origin") is None

    # no snapshot on origin visit so None
    origin = Origin.from_dict(data.origin)
    swh_storage.origin_add_one(origin)
    swh_storage.origin_visit_add(
        [OriginVisit(origin=origin.url, date=data.date_visit1, type=data.type_visit1,)]
    )
    assert snapshot_get_latest(swh_storage, origin.url) is None

    ov1 = swh_storage.origin_visit_get_latest(origin.url)
    assert ov1 is not None
    visit_id = ov1["visit"]

    # visit references a snapshot but the snapshot does not exist in backend for some
    # reason
    complete_snapshot = Snapshot.from_dict(data.complete_snapshot)
    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=visit_id,
                date=data.date_visit2,
                status="partial",
                snapshot=complete_snapshot.id,
            )
        ]
    )
    # so we do not find it
    assert snapshot_get_latest(swh_storage, origin.url) is None
    assert snapshot_get_latest(swh_storage, origin.url, branches_count=1) is None


def test_snapshot_get_latest(swh_storage):
    origin = Origin.from_dict(data.origin)
    swh_storage.origin_add_one(origin)

    visit1 = OriginVisit(
        origin=origin.url, date=data.date_visit1, type=data.type_visit1,
    )
    swh_storage.origin_visit_add([visit1])
    ov1 = swh_storage.origin_visit_get_latest(origin.url)
    visit_id = ov1["visit"]

    # Add snapshot to visit1, latest snapshot = visit 1 snapshot
    complete_snapshot = Snapshot.from_dict(data.complete_snapshot)
    swh_storage.snapshot_add([complete_snapshot])

    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=visit_id,
                date=data.date_visit2,
                status="partial",
                snapshot=None,
            )
        ]
    )
    assert data.date_visit1 < data.date_visit2

    # no snapshot associated to the visit, so None
    actual_snapshot = snapshot_get_latest(
        swh_storage, origin.url, allowed_statuses=["partial"]
    )
    assert actual_snapshot is None

    date_now = now()
    assert data.date_visit2 < date_now
    swh_storage.origin_visit_status_add(
        [
            OriginVisitStatus(
                origin=origin.url,
                visit=visit_id,
                date=date_now,
                status="full",
                snapshot=complete_snapshot.id,
            )
        ]
    )

    swh_storage.origin_visit_add(
        [OriginVisit(origin=origin.url, date=now(), type=data.type_visit1,)]
    )

    actual_snapshot = snapshot_get_latest(swh_storage, origin.url)
    assert actual_snapshot is not None
    assert actual_snapshot == complete_snapshot

    actual_snapshot = snapshot_get_latest(swh_storage, origin.url, branches_count=1)
    assert actual_snapshot is not None
    assert actual_snapshot.id == complete_snapshot.id
    assert len(actual_snapshot.branches.values()) == 1

    with pytest.raises(ValueError, match="branches_count must be a positive integer"):
        snapshot_get_latest(swh_storage, origin.url, branches_count="something-wrong")



# Copyright (C) 2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import call

import psycopg2
import pytest

from swh.model.model import (
    Content,
    Directory,
    Release,
    Revision,
    Snapshot,
    SkippedContent,
    Origin,
    OriginVisit,
)

from swh.storage import get_storage
from swh.storage.exc import HashCollision, StorageArgumentException

from .storage_data import date_visit1


@pytest.fixture
def monkeypatch_sleep(monkeypatch, swh_storage):
    """In test context, we don't want to wait, make test faster

    """
    from swh.storage.retry import RetryingProxyStorage

    for method_name, method in RetryingProxyStorage.__dict__.items():
        if "_add" in method_name or "_update" in method_name:
            monkeypatch.setattr(method.retry, "sleep", lambda x: None)

    return monkeypatch


@pytest.fixture
def fake_hash_collision(sample_data):
    return HashCollision("sha1", "38762cf7f55934b34d179ae6a4c80cadccbb7f0a", [])


@pytest.fixture
def swh_storage():
    storage_config = {
        "cls": "pipeline",
        "steps": [{"cls": "validate"}, {"cls": "retry"}, {"cls": "memory"},],
    }
    return get_storage(**storage_config)


def test_retrying_proxy_storage_content_add(swh_storage, sample_data):
    """Standard content_add works as before

    """
    sample_content = sample_data["content"][0]

    content = next(swh_storage.content_get([sample_content["sha1"]]))
    assert not content

    s = swh_storage.content_add([sample_content])
    assert s == {
        "content:add": 1,
        "content:add:bytes": sample_content["length"],
    }

    content = next(swh_storage.content_get([sample_content["sha1"]]))
    assert content["sha1"] == sample_content["sha1"]


def test_retrying_proxy_storage_content_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision,
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.content_add")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("content already inserted"),
        # ok then!
        {"content:add": 1},
    ]

    sample_content = sample_data["content"][0]

    content = next(swh_storage.content_get([sample_content["sha1"]]))
    assert not content

    s = swh_storage.content_add([sample_content])
    assert s == {"content:add": 1}

    mock_memory.assert_has_calls(
        [
            call([Content.from_dict(sample_content)]),
            call([Content.from_dict(sample_content)]),
            call([Content.from_dict(sample_content)]),
        ]
    )


def test_retrying_proxy_swh_storage_content_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.content_add")
    mock_memory.side_effect = StorageArgumentException("Refuse to add content always!")

    sample_content = sample_data["content"][0]

    content = next(swh_storage.content_get([sample_content["sha1"]]))
    assert not content

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.content_add([sample_content])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_content_add_metadata(swh_storage, sample_data):
    """Standard content_add_metadata works as before

    """
    sample_content = sample_data["content_metadata"][0]

    pk = sample_content["sha1"]
    content_metadata = swh_storage.content_get_metadata([pk])
    assert not content_metadata[pk]

    s = swh_storage.content_add_metadata([sample_content])
    assert s == {
        "content:add": 1,
    }

    content_metadata = swh_storage.content_get_metadata([pk])
    assert len(content_metadata[pk]) == 1
    assert content_metadata[pk][0]["sha1"] == pk


def test_retrying_proxy_storage_content_add_metadata_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.content_add_metadata"
    )
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("content_metadata already inserted"),
        # ok then!
        {"content:add": 1},
    ]

    sample_content = sample_data["content_metadata"][0]

    s = swh_storage.content_add_metadata([sample_content])
    assert s == {"content:add": 1}

    mock_memory.assert_has_calls(
        [
            call([Content.from_dict(sample_content)]),
            call([Content.from_dict(sample_content)]),
            call([Content.from_dict(sample_content)]),
        ]
    )


def test_retrying_proxy_swh_storage_content_add_metadata_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.content_add_metadata"
    )
    mock_memory.side_effect = StorageArgumentException(
        "Refuse to add content_metadata!"
    )

    sample_content = sample_data["content_metadata"][0]
    pk = sample_content["sha1"]

    content_metadata = swh_storage.content_get_metadata([pk])
    assert not content_metadata[pk]

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.content_add_metadata([sample_content])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_skipped_content_add(swh_storage, sample_data):
    """Standard skipped_content_add works as before

    """
    sample_content = sample_data["skipped_content"][0]

    skipped_contents = list(swh_storage.skipped_content_missing([sample_content]))
    assert len(skipped_contents) == 1

    s = swh_storage.skipped_content_add([sample_content])
    assert s == {
        "skipped_content:add": 1,
    }

    skipped_content = list(swh_storage.skipped_content_missing([sample_content]))
    assert len(skipped_content) == 0


def test_retrying_proxy_storage_skipped_content_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.skipped_content_add"
    )
    mock_memory.side_effect = [
        # 1st & 2nd try goes ko
        fake_hash_collision,
        psycopg2.IntegrityError("skipped_content already inserted"),
        # ok then!
        {"skipped_content:add": 1},
    ]

    sample_content = sample_data["skipped_content"][0]

    s = swh_storage.skipped_content_add([sample_content])
    assert s == {"skipped_content:add": 1}

    mock_memory.assert_has_calls(
        [
            call([SkippedContent.from_dict(sample_content)]),
            call([SkippedContent.from_dict(sample_content)]),
            call([SkippedContent.from_dict(sample_content)]),
        ]
    )


def test_retrying_proxy_swh_storage_skipped_content_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.skipped_content_add"
    )
    mock_memory.side_effect = StorageArgumentException(
        "Refuse to add content_metadata!"
    )

    sample_content = sample_data["skipped_content"][0]

    skipped_contents = list(swh_storage.skipped_content_missing([sample_content]))
    assert len(skipped_contents) == 1

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.skipped_content_add([sample_content])

    skipped_contents = list(swh_storage.skipped_content_missing([sample_content]))
    assert len(skipped_contents) == 1

    assert mock_memory.call_count == 1


def test_retrying_proxy_swh_storage_origin_add_one(swh_storage, sample_data):
    """Standard origin_add_one works as before

    """
    sample_origin = sample_data["origin"][0]

    origin = swh_storage.origin_get(sample_origin)
    assert not origin

    r = swh_storage.origin_add_one(sample_origin)
    assert r == sample_origin["url"]

    origin = swh_storage.origin_get(sample_origin)
    assert origin["url"] == sample_origin["url"]


def test_retrying_proxy_swh_storage_origin_add_one_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    sample_origin = sample_data["origin"][1]
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.origin_add_one")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("origin already inserted"),
        # ok then!
        sample_origin["url"],
    ]

    origin = swh_storage.origin_get(sample_origin)
    assert not origin

    r = swh_storage.origin_add_one(sample_origin)
    assert r == sample_origin["url"]

    mock_memory.assert_has_calls(
        [
            call(Origin.from_dict(sample_origin)),
            call(Origin.from_dict(sample_origin)),
            call(Origin.from_dict(sample_origin)),
        ]
    )


def test_retrying_proxy_swh_storage_origin_add_one_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.origin_add_one")
    mock_memory.side_effect = StorageArgumentException("Refuse to add origin always!")

    sample_origin = sample_data["origin"][0]

    origin = swh_storage.origin_get(sample_origin)
    assert not origin

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.origin_add_one(sample_origin)

    assert mock_memory.call_count == 1


def test_retrying_proxy_swh_storage_origin_visit_add(swh_storage, sample_data):
    """Standard origin_visit_add works as before

    """
    sample_origin = sample_data["origin"][0]
    origin_url = swh_storage.origin_add_one(sample_origin)

    origin = list(swh_storage.origin_visit_get(origin_url))
    assert not origin

    visit = OriginVisit(origin=origin_url, date=date_visit1, type="hg")
    origin_visit = swh_storage.origin_visit_add([visit])[0]
    assert origin_visit.origin == origin_url
    assert isinstance(origin_visit.visit, int)

    origin_visit = next(swh_storage.origin_visit_get(origin_url))
    assert origin_visit["origin"] == origin_url
    assert isinstance(origin_visit["visit"], int)


def test_retrying_proxy_swh_storage_origin_visit_add_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    sample_origin = sample_data["origin"][1]
    origin_url = swh_storage.origin_add_one(sample_origin)

    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.origin_visit_add")
    visit = OriginVisit(origin=origin_url, date=date_visit1, type="git")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("origin already inserted"),
        # ok then!
        [visit],
    ]

    origin = list(swh_storage.origin_visit_get(origin_url))
    assert not origin

    r = swh_storage.origin_visit_add([visit])
    assert r == [visit]

    mock_memory.assert_has_calls(
        [call([visit]), call([visit]), call([visit]),]
    )


def test_retrying_proxy_swh_storage_origin_visit_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.origin_visit_add")
    mock_memory.side_effect = StorageArgumentException("Refuse to add origin always!")

    origin_url = sample_data["origin"][0]["url"]

    origin = list(swh_storage.origin_visit_get(origin_url))
    assert not origin

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        visit = OriginVisit(origin=origin_url, date=date_visit1, type="svn",)
        swh_storage.origin_visit_add([visit])

    mock_memory.assert_has_calls(
        [call([visit]),]
    )


def test_retrying_proxy_storage_metadata_fetcher_add(swh_storage, sample_data):
    """Standard metadata_fetcher_add works as before

    """
    fetcher = sample_data["fetcher"][0]

    metadata_fetcher = swh_storage.metadata_fetcher_get(
        fetcher["name"], fetcher["version"]
    )
    assert not metadata_fetcher

    swh_storage.metadata_fetcher_add(**fetcher)

    actual_fetcher = swh_storage.metadata_fetcher_get(
        fetcher["name"], fetcher["version"]
    )
    assert actual_fetcher == fetcher


def test_retrying_proxy_storage_metadata_fetcher_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    fetcher = sample_data["fetcher"][0]
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.metadata_fetcher_add"
    )
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("metadata_fetcher already inserted"),
        # ok then!
        [fetcher],
    ]

    actual_fetcher = swh_storage.metadata_fetcher_get(
        fetcher["name"], fetcher["version"]
    )
    assert not actual_fetcher

    swh_storage.metadata_fetcher_add(**fetcher)

    mock_memory.assert_has_calls(
        [
            call(fetcher["name"], fetcher["version"], fetcher["metadata"]),
            call(fetcher["name"], fetcher["version"], fetcher["metadata"]),
            call(fetcher["name"], fetcher["version"], fetcher["metadata"]),
        ]
    )


def test_retrying_proxy_swh_storage_metadata_fetcher_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.metadata_fetcher_add"
    )
    mock_memory.side_effect = StorageArgumentException(
        "Refuse to add metadata_fetcher always!"
    )

    fetcher = sample_data["fetcher"][0]

    actual_fetcher = swh_storage.metadata_fetcher_get(
        fetcher["name"], fetcher["version"]
    )
    assert not actual_fetcher

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.metadata_fetcher_add(**fetcher)

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_metadata_authority_add(swh_storage, sample_data):
    """Standard metadata_authority_add works as before

    """
    authority = sample_data["authority"][0]

    assert not swh_storage.metadata_authority_get(authority["type"], authority["url"])

    swh_storage.metadata_authority_add(**authority)

    actual_authority = swh_storage.metadata_authority_get(
        authority["type"], authority["url"]
    )
    assert actual_authority == authority


def test_retrying_proxy_storage_metadata_authority_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    authority = sample_data["authority"][0]

    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.metadata_authority_add"
    )
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("foo bar"),
        # ok then!
        None,
    ]

    assert not swh_storage.metadata_authority_get(authority["type"], authority["url"])

    swh_storage.metadata_authority_add(**authority)

    authority_arg_names = ("type", "url", "metadata")
    authority_args = [authority[key] for key in authority_arg_names]
    mock_memory.assert_has_calls(
        [call(*authority_args), call(*authority_args), call(*authority_args),]
    )


def test_retrying_proxy_swh_storage_metadata_authority_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.metadata_authority_add"
    )
    mock_memory.side_effect = StorageArgumentException(
        "Refuse to add authority_id always!"
    )

    authority = sample_data["authority"][0]

    swh_storage.metadata_authority_get(authority["type"], authority["url"])

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.metadata_authority_add(**authority)

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_origin_metadata_add(swh_storage, sample_data):
    """Standard origin_metadata_add works as before

    """
    ori_meta = sample_data["origin_metadata"][0]
    swh_storage.origin_add_one({"url": ori_meta["origin_url"]})
    swh_storage.metadata_authority_add(**sample_data["authority"][0])
    swh_storage.metadata_fetcher_add(**sample_data["fetcher"][0])

    origin_metadata = swh_storage.origin_metadata_get(
        ori_meta["origin_url"], ori_meta["authority"]
    )
    assert origin_metadata["next_page_token"] is None
    assert not origin_metadata["results"]

    swh_storage.origin_metadata_add(**ori_meta)

    origin_metadata = swh_storage.origin_metadata_get(
        ori_meta["origin_url"], ori_meta["authority"]
    )
    assert origin_metadata


def test_retrying_proxy_storage_origin_metadata_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    ori_meta = sample_data["origin_metadata"][0]
    swh_storage.origin_add_one({"url": ori_meta["origin_url"]})
    swh_storage.metadata_authority_add(**sample_data["authority"][0])
    swh_storage.metadata_fetcher_add(**sample_data["fetcher"][0])
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.origin_metadata_add"
    )

    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("foo bar"),
        # ok then!
        None,
    ]

    # No exception raised as insertion finally came through
    swh_storage.origin_metadata_add(**ori_meta)

    mock_memory.assert_has_calls(
        [  # 3 calls, as long as error raised
            call(
                ori_meta["origin_url"],
                ori_meta["discovery_date"],
                ori_meta["authority"],
                ori_meta["fetcher"],
                ori_meta["format"],
                ori_meta["metadata"],
            ),
            call(
                ori_meta["origin_url"],
                ori_meta["discovery_date"],
                ori_meta["authority"],
                ori_meta["fetcher"],
                ori_meta["format"],
                ori_meta["metadata"],
            ),
            call(
                ori_meta["origin_url"],
                ori_meta["discovery_date"],
                ori_meta["authority"],
                ori_meta["fetcher"],
                ori_meta["format"],
                ori_meta["metadata"],
            ),
        ]
    )


def test_retrying_proxy_swh_storage_origin_metadata_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch(
        "swh.storage.in_memory.InMemoryStorage.origin_metadata_add"
    )
    mock_memory.side_effect = StorageArgumentException("Refuse to add always!")

    ori_meta = sample_data["origin_metadata"][0]
    swh_storage.origin_add_one({"url": ori_meta["origin_url"]})

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.origin_metadata_add(**ori_meta)

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_directory_add(swh_storage, sample_data):
    """Standard directory_add works as before

    """
    sample_dir = sample_data["directory"][0]

    directory = swh_storage.directory_get_random()  # no directory
    assert not directory

    s = swh_storage.directory_add([sample_dir])
    assert s == {
        "directory:add": 1,
    }

    directory_id = swh_storage.directory_get_random()  # only 1
    assert directory_id == sample_dir["id"]


def test_retrying_proxy_storage_directory_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.directory_add")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("directory already inserted"),
        # ok then!
        {"directory:add": 1},
    ]

    sample_dir = sample_data["directory"][1]

    directory_id = swh_storage.directory_get_random()  # no directory
    assert not directory_id

    s = swh_storage.directory_add([sample_dir])
    assert s == {
        "directory:add": 1,
    }

    mock_memory.assert_has_calls(
        [
            call([Directory.from_dict(sample_dir)]),
            call([Directory.from_dict(sample_dir)]),
            call([Directory.from_dict(sample_dir)]),
        ]
    )


def test_retrying_proxy_swh_storage_directory_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.directory_add")
    mock_memory.side_effect = StorageArgumentException(
        "Refuse to add directory always!"
    )

    sample_dir = sample_data["directory"][0]

    directory_id = swh_storage.directory_get_random()  # no directory
    assert not directory_id

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.directory_add([sample_dir])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_revision_add(swh_storage, sample_data):
    """Standard revision_add works as before

    """
    sample_rev = sample_data["revision"][0]

    revision = next(swh_storage.revision_get([sample_rev["id"]]))
    assert not revision

    s = swh_storage.revision_add([sample_rev])
    assert s == {
        "revision:add": 1,
    }

    revision = next(swh_storage.revision_get([sample_rev["id"]]))
    assert revision["id"] == sample_rev["id"]


def test_retrying_proxy_storage_revision_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.revision_add")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("revision already inserted"),
        # ok then!
        {"revision:add": 1},
    ]

    sample_rev = sample_data["revision"][0]

    revision = next(swh_storage.revision_get([sample_rev["id"]]))
    assert not revision

    s = swh_storage.revision_add([sample_rev])
    assert s == {
        "revision:add": 1,
    }

    mock_memory.assert_has_calls(
        [
            call([Revision.from_dict(sample_rev)]),
            call([Revision.from_dict(sample_rev)]),
            call([Revision.from_dict(sample_rev)]),
        ]
    )


def test_retrying_proxy_swh_storage_revision_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.revision_add")
    mock_memory.side_effect = StorageArgumentException("Refuse to add revision always!")

    sample_rev = sample_data["revision"][0]

    revision = next(swh_storage.revision_get([sample_rev["id"]]))
    assert not revision

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.revision_add([sample_rev])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_release_add(swh_storage, sample_data):
    """Standard release_add works as before

    """
    sample_rel = sample_data["release"][0]

    release = next(swh_storage.release_get([sample_rel["id"]]))
    assert not release

    s = swh_storage.release_add([sample_rel])
    assert s == {
        "release:add": 1,
    }

    release = next(swh_storage.release_get([sample_rel["id"]]))
    assert release["id"] == sample_rel["id"]


def test_retrying_proxy_storage_release_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.release_add")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("release already inserted"),
        # ok then!
        {"release:add": 1},
    ]

    sample_rel = sample_data["release"][0]

    release = next(swh_storage.release_get([sample_rel["id"]]))
    assert not release

    s = swh_storage.release_add([sample_rel])
    assert s == {
        "release:add": 1,
    }

    mock_memory.assert_has_calls(
        [
            call([Release.from_dict(sample_rel)]),
            call([Release.from_dict(sample_rel)]),
            call([Release.from_dict(sample_rel)]),
        ]
    )


def test_retrying_proxy_swh_storage_release_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.release_add")
    mock_memory.side_effect = StorageArgumentException("Refuse to add release always!")

    sample_rel = sample_data["release"][0]

    release = next(swh_storage.release_get([sample_rel["id"]]))
    assert not release

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.release_add([sample_rel])

    assert mock_memory.call_count == 1


def test_retrying_proxy_storage_snapshot_add(swh_storage, sample_data):
    """Standard snapshot_add works as before

    """
    sample_snap = sample_data["snapshot"][0]

    snapshot = swh_storage.snapshot_get(sample_snap["id"])
    assert not snapshot

    s = swh_storage.snapshot_add([sample_snap])
    assert s == {
        "snapshot:add": 1,
    }

    snapshot = swh_storage.snapshot_get(sample_snap["id"])
    assert snapshot["id"] == sample_snap["id"]


def test_retrying_proxy_storage_snapshot_add_with_retry(
    monkeypatch_sleep, swh_storage, sample_data, mocker, fake_hash_collision
):
    """Multiple retries for hash collision and psycopg2 error but finally ok

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.snapshot_add")
    mock_memory.side_effect = [
        # first try goes ko
        fake_hash_collision,
        # second try goes ko
        psycopg2.IntegrityError("snapshot already inserted"),
        # ok then!
        {"snapshot:add": 1},
    ]

    sample_snap = sample_data["snapshot"][0]

    snapshot = swh_storage.snapshot_get(sample_snap["id"])
    assert not snapshot

    s = swh_storage.snapshot_add([sample_snap])
    assert s == {
        "snapshot:add": 1,
    }

    mock_memory.assert_has_calls(
        [
            call([Snapshot.from_dict(sample_snap)]),
            call([Snapshot.from_dict(sample_snap)]),
            call([Snapshot.from_dict(sample_snap)]),
        ]
    )


def test_retrying_proxy_swh_storage_snapshot_add_failure(
    swh_storage, sample_data, mocker
):
    """Unfiltered errors are raising without retry

    """
    mock_memory = mocker.patch("swh.storage.in_memory.InMemoryStorage.snapshot_add")
    mock_memory.side_effect = StorageArgumentException("Refuse to add snapshot always!")

    sample_snap = sample_data["snapshot"][0]

    snapshot = swh_storage.snapshot_get(sample_snap["id"])
    assert not snapshot

    with pytest.raises(StorageArgumentException, match="Refuse to add"):
        swh_storage.snapshot_add([sample_snap])

    assert mock_memory.call_count == 1



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.storage import get_storage


def get_storage_with_buffer_config(**buffer_config):
    storage_config = {
        "cls": "pipeline",
        "steps": [
            {"cls": "validate"},
            {"cls": "buffer", **buffer_config},
            {"cls": "memory"},
        ],
    }

    return get_storage(**storage_config)


def test_buffering_proxy_storage_content_threshold_not_hit(sample_data):
    contents = sample_data["content"]
    storage = get_storage_with_buffer_config(min_batch_size={"content": 10,})
    s = storage.content_add([contents[0], contents[1]])
    assert s == {}

    # contents have not been written to storage
    missing_contents = storage.content_missing([contents[0], contents[1]])
    assert set(missing_contents) == set([contents[0]["sha1"], contents[1]["sha1"]])

    s = storage.flush()
    assert s == {
        "content:add": 1 + 1,
        "content:add:bytes": contents[0]["length"] + contents[1]["length"],
    }

    missing_contents = storage.content_missing([contents[0], contents[1]])
    assert list(missing_contents) == []


def test_buffering_proxy_storage_content_threshold_nb_hit(sample_data):
    contents = sample_data["content"]
    storage = get_storage_with_buffer_config(min_batch_size={"content": 1,})

    s = storage.content_add([contents[0]])
    assert s == {
        "content:add": 1,
        "content:add:bytes": contents[0]["length"],
    }

    missing_contents = storage.content_missing([contents[0]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_content_deduplicate(sample_data):
    contents = sample_data["content"]
    storage = get_storage_with_buffer_config(min_batch_size={"content": 2,})

    s = storage.content_add([contents[0], contents[0]])
    assert s == {}

    s = storage.content_add([contents[0]])
    assert s == {}

    s = storage.content_add([contents[1]])
    assert s == {
        "content:add": 1 + 1,
        "content:add:bytes": contents[0]["length"] + contents[1]["length"],
    }

    missing_contents = storage.content_missing([contents[0], contents[1]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_content_threshold_bytes_hit(sample_data):
    contents = sample_data["content"]
    content_bytes_min_batch_size = 2
    storage = get_storage_with_buffer_config(
        min_batch_size={"content": 10, "content_bytes": content_bytes_min_batch_size,}
    )

    assert contents[0]["length"] > content_bytes_min_batch_size

    s = storage.content_add([contents[0]])
    assert s == {
        "content:add": 1,
        "content:add:bytes": contents[0]["length"],
    }

    missing_contents = storage.content_missing([contents[0]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_skipped_content_threshold_not_hit(sample_data):
    contents = sample_data["skipped_content"]
    storage = get_storage_with_buffer_config(min_batch_size={"skipped_content": 10,})
    s = storage.skipped_content_add([contents[0], contents[1]])
    assert s == {}

    # contents have not been written to storage
    missing_contents = storage.skipped_content_missing([contents[0], contents[1]])
    assert {c["sha1"] for c in missing_contents} == {c["sha1"] for c in contents}

    s = storage.flush()
    assert s == {"skipped_content:add": 1 + 1}

    missing_contents = storage.skipped_content_missing([contents[0], contents[1]])
    assert list(missing_contents) == []


def test_buffering_proxy_storage_skipped_content_threshold_nb_hit(sample_data):
    contents = sample_data["skipped_content"]
    storage = get_storage_with_buffer_config(min_batch_size={"skipped_content": 1,})

    s = storage.skipped_content_add([contents[0]])
    assert s == {"skipped_content:add": 1}

    missing_contents = storage.skipped_content_missing([contents[0]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_skipped_content_deduplicate(sample_data):
    contents = sample_data["skipped_content"]
    storage = get_storage_with_buffer_config(min_batch_size={"skipped_content": 2,})

    s = storage.skipped_content_add([contents[0], contents[0]])
    assert s == {}

    s = storage.skipped_content_add([contents[0]])
    assert s == {}

    s = storage.skipped_content_add([contents[1]])
    assert s == {
        "skipped_content:add": 1 + 1,
    }

    missing_contents = storage.skipped_content_missing([contents[0], contents[1]])
    assert list(missing_contents) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_directory_threshold_not_hit(sample_data):
    directories = sample_data["directory"]
    storage = get_storage_with_buffer_config(min_batch_size={"directory": 10,})
    s = storage.directory_add([directories[0]])
    assert s == {}

    directory_id = directories[0]["id"]
    missing_directories = storage.directory_missing([directory_id])
    assert list(missing_directories) == [directory_id]

    s = storage.flush()
    assert s == {
        "directory:add": 1,
    }

    missing_directories = storage.directory_missing([directory_id])
    assert list(missing_directories) == []


def test_buffering_proxy_storage_directory_threshold_hit(sample_data):
    directories = sample_data["directory"]
    storage = get_storage_with_buffer_config(min_batch_size={"directory": 1,})
    s = storage.directory_add([directories[0]])
    assert s == {
        "directory:add": 1,
    }

    missing_directories = storage.directory_missing([directories[0]["id"]])
    assert list(missing_directories) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_directory_deduplicate(sample_data):
    directories = sample_data["directory"]
    storage = get_storage_with_buffer_config(min_batch_size={"directory": 2,})

    s = storage.directory_add([directories[0], directories[0]])
    assert s == {}

    s = storage.directory_add([directories[0]])
    assert s == {}

    s = storage.directory_add([directories[1]])
    assert s == {
        "directory:add": 1 + 1,
    }

    missing_directories = storage.directory_missing(
        [directories[0]["id"], directories[1]["id"]]
    )
    assert list(missing_directories) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_revision_threshold_not_hit(sample_data):
    revisions = sample_data["revision"]
    storage = get_storage_with_buffer_config(min_batch_size={"revision": 10,})
    s = storage.revision_add([revisions[0]])
    assert s == {}

    revision_id = revisions[0]["id"]
    missing_revisions = storage.revision_missing([revision_id])
    assert list(missing_revisions) == [revision_id]

    s = storage.flush()
    assert s == {
        "revision:add": 1,
    }

    missing_revisions = storage.revision_missing([revision_id])
    assert list(missing_revisions) == []


def test_buffering_proxy_storage_revision_threshold_hit(sample_data):
    revisions = sample_data["revision"]
    storage = get_storage_with_buffer_config(min_batch_size={"revision": 1,})
    s = storage.revision_add([revisions[0]])
    assert s == {
        "revision:add": 1,
    }

    missing_revisions = storage.revision_missing([revisions[0]["id"]])
    assert list(missing_revisions) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_revision_deduplicate(sample_data):
    revisions = sample_data["revision"]
    storage = get_storage_with_buffer_config(min_batch_size={"revision": 2,})

    s = storage.revision_add([revisions[0], revisions[0]])
    assert s == {}

    s = storage.revision_add([revisions[0]])
    assert s == {}

    s = storage.revision_add([revisions[1]])
    assert s == {
        "revision:add": 1 + 1,
    }

    missing_revisions = storage.revision_missing(
        [revisions[0]["id"], revisions[1]["id"]]
    )
    assert list(missing_revisions) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_release_threshold_not_hit(sample_data):
    releases = sample_data["release"]
    threshold = 10

    assert len(releases) < threshold
    storage = get_storage_with_buffer_config(
        min_batch_size={"release": threshold,}  # configuration set
    )
    s = storage.release_add(releases)
    assert s == {}

    release_ids = [r["id"] for r in releases]
    missing_releases = storage.release_missing(release_ids)
    assert list(missing_releases) == release_ids

    s = storage.flush()
    assert s == {
        "release:add": len(releases),
    }

    missing_releases = storage.release_missing(release_ids)
    assert list(missing_releases) == []


def test_buffering_proxy_storage_release_threshold_hit(sample_data):
    releases = sample_data["release"]
    threshold = 2
    assert len(releases) > threshold

    storage = get_storage_with_buffer_config(
        min_batch_size={"release": threshold,}  # configuration set
    )

    s = storage.release_add(releases)
    assert s == {
        "release:add": len(releases),
    }

    release_ids = [r["id"] for r in releases]
    missing_releases = storage.release_missing(release_ids)
    assert list(missing_releases) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_release_deduplicate(sample_data):
    releases = sample_data["release"]
    storage = get_storage_with_buffer_config(min_batch_size={"release": 2,})

    s = storage.release_add([releases[0], releases[0]])
    assert s == {}

    s = storage.release_add([releases[0]])
    assert s == {}

    s = storage.release_add([releases[1]])
    assert s == {
        "release:add": 1 + 1,
    }

    missing_releases = storage.release_missing([releases[0]["id"], releases[1]["id"]])
    assert list(missing_releases) == []

    s = storage.flush()
    assert s == {}


def test_buffering_proxy_storage_clear(sample_data):
    """Clear operation on buffer

    """
    threshold = 10
    contents = sample_data["content"]
    assert 0 < len(contents) < threshold
    skipped_contents = sample_data["skipped_content"]
    assert 0 < len(skipped_contents) < threshold
    directories = sample_data["directory"]
    assert 0 < len(directories) < threshold
    revisions = sample_data["revision"]
    assert 0 < len(revisions) < threshold
    releases = sample_data["release"]
    assert 0 < len(releases) < threshold

    storage = get_storage_with_buffer_config(
        min_batch_size={
            "content": threshold,
            "skipped_content": threshold,
            "directory": threshold,
            "revision": threshold,
            "release": threshold,
        }
    )

    s = storage.content_add(contents)
    assert s == {}
    s = storage.skipped_content_add(skipped_contents)
    assert s == {}
    s = storage.directory_add(directories)
    assert s == {}
    s = storage.revision_add(revisions)
    assert s == {}
    s = storage.release_add(releases)
    assert s == {}

    assert len(storage._objects["content"]) == len(contents)
    assert len(storage._objects["skipped_content"]) == len(skipped_contents)
    assert len(storage._objects["directory"]) == len(directories)
    assert len(storage._objects["revision"]) == len(revisions)
    assert len(storage._objects["release"]) == len(releases)

    # clear only content from the buffer
    s = storage.clear_buffers(["content"])
    assert s is None

    # specific clear operation on specific object type content only touched
    # them
    assert len(storage._objects["content"]) == 0
    assert len(storage._objects["skipped_content"]) == len(skipped_contents)
    assert len(storage._objects["directory"]) == len(directories)
    assert len(storage._objects["revision"]) == len(revisions)
    assert len(storage._objects["release"]) == len(releases)

    # clear current buffer from all object types
    s = storage.clear_buffers()
    assert s is None

    assert len(storage._objects["content"]) == 0
    assert len(storage._objects["skipped_content"]) == 0
    assert len(storage._objects["directory"]) == 0
    assert len(storage._objects["revision"]) == 0
    assert len(storage._objects["release"]) == 0
